import sys
from scientisst.scientisst import AX1, AX2
from sense_src.thread_builder import ThreadBuilder
from datetime import datetime


class FileWriter(ThreadBuilder):
    def __init__(
        self, filename, address, fs, channels, mv, api_version, firmware_version
    ):
        super().__init__()
        self.filename = filename
        self.mv = mv
        self.channels = channels
        self.metadata = self.__get_metadata(
            address, fs, channels, api_version, firmware_version
        )

    def start(self):
        self.__init_file()
        super().start()

    def stop(self):
        super().stop()
        if self.f:
            self.f.close()

    def thread_method(self, frames):
        self.f.write("\n".join(map(str, frames)) + "\n")

    def __init_file(
        self,
    ):
        self.f = open(self.filename, "w")
        sys.stdout.write("Saving data to {}\n".format(self.filename))

        header = "\t".join(self.metadata["Header"])

        self.f.write("#{}\n".format(self.metadata))
        self.f.write("#{}\n".format(header))

    def __get_metadata(self, address, fs, channels, api_version, firmware_version):
        timestamp = datetime.now()
        metadata = {
            "API version": api_version,
            "Channels": channels,
            "Channels labels": get_channel_labels(channels, self.mv),
            "Device": address,
            "Firmware version": firmware_version,
            "Header": get_header(channels, self.mv),
            "Resolution (bits)": [4, 1, 1, 1, 1] + self.__get_channel_resolutions(),
            "Sampling rate (Hz)": fs,
            "Timestamp": timestamp.timestamp(),
            "ISO 8601": timestamp.isoformat(),
        }
        if self.mv:
            metadata["Channels indexes raw"] = list(
                map(lambda x: (x - 1) * 2 + 5, channels)
            )
            metadata["Channels indexes mV"] = list(
                map(lambda x: (x - 1) * 2 + 6, channels)
            )
        else:
            metadata["Channels indexes"] = list(map(lambda x: x + 5, channels))

        sorted_metadata = {}
        for key in sorted(metadata):
            sorted_metadata[key] = metadata[key]

        return sorted_metadata

    def __get_channel_resolutions(self):
        channel_resolutions = []
        for ch in self.channels:
            if ch == AX1 or ch == AX2:
                channel_resolutions += [24]
            else:
                channel_resolutions += [12]
        return channel_resolutions

    def __get_channel_resolutions_mv(self):
        channel_resolutions = []
        for ch in self.channels:
            if ch == AX1 or ch == AX2:
                channel_resolutions += [0.4]
            else:
                channel_resolutions += [0.8]
        return channel_resolutions


def get_channel_labels(channels, mv):
    channel_labels = []
    for ch in channels:
        if not mv:
            if ch == AX1 or ch == AX2:
                channel_labels += ["AX{}".format(ch)]
            else:
                channel_labels += ["AI{}".format(ch)]
        else:
            if ch == AX1 or ch == AX2:
                channel_labels += ["AX{}_raw".format(ch)]
                channel_labels += ["AX{}_mv".format(ch)]
            else:
                channel_labels += ["AI{}_raw".format(ch)]
                channel_labels += ["AI{}_mv".format(ch)]
    return channel_labels


def get_header(channels, mv):
    header = ["NSeq", "I1", "I2", "O1", "O2"]
    header += get_channel_labels(channels, mv)
    return header



import sys


class DevicePicker:
    def select_device(self):
        options, labels = self.__get_device_options()
        if len(options) > 0:
            sys.stdout.write("ScientISST devices:\n")
            label_index = 1
            for label in labels:
                sys.stdout.write("[{}] {}\n".format(label_index, label))
                label_index += 1
            selected_index = 0
            while selected_index == 0:
                user_input = input("Connect to: ")
                try:
                    selected_index = int(user_input)
                    if selected_index > len(options):
                        selected_index = 0
                        raise ValueError()
                except ValueError:
                    sys.stderr.write('"{}" is not a valid index\n'.format(user_input))
            return options[selected_index - 1]

    def __get_device_options(self):
        if sys.platform == "linux":
            options = self.__get_linux_bth_devices()
            return list(map(lambda option: option["addr"], options)), list(
                map(
                    lambda option: "{} - {}".format(
                        option["name"] if "name" in option else "unnamed",
                        option["addr"],
                    ),
                    options,
                )
            )
        else:
            import serial.tools.list_ports

            ports = serial.tools.list_ports.comports()
            options = []
            labels = []
            for port, desc, hwid in sorted(ports):
                if "scientisst" in port.lower():
                    options += [port]
                    label = ""
                    if desc != "n/a":
                        label += "{} - ".format(desc)
                    labels += [label + port]
            return options, labels

    def __get_linux_bth_devices(self):

        import pydbus

        bt_devices = {}

        bus = pydbus.SystemBus()
        mngr = bus.get("org.bluez", "/")

        mngd_objs = mngr.GetManagedObjects()
        for path in mngd_objs:
            addr = mngd_objs[path].get("org.bluez.Device1", {}).get("Address")
            name = mngd_objs[path].get("org.bluez.Device1", {}).get("Name")

            if name and "scientisst" in name.lower() and addr not in bt_devices:
                bt_devices[addr] = {
                    "name": name,
                    "addr": addr,
                }

        return bt_devices.values()



#!/usr/bin/python

"""
sense.py
"""

VERSION = "0.1.7"

import sys
from scientisst import *
from threading import Timer
from threading import Event
from sense_src.arg_parser import ArgParser
from sense_src.custom_script import get_custom_script, CustomScript
from sense_src.device_picker import DevicePicker
from sense_src.file_writer import *


def run_scheduled_task(duration, stop_event):
    def stop(stop_event):
        stop_event.set()

    timer = Timer(duration, stop, [stop_event])
    timer.start()
    return timer


def main():
    arg_parser = ArgParser()
    args = arg_parser.args

    if args.version:
        sys.stdout.write("sense.py version {}\n".format(VERSION))
        sys.exit(0)

    if args.address:
        address = args.address
    else:
        address = DevicePicker().select_device()
        if not address:
            arg_parser.error("No paired device found")

    args.channels = sorted(map(int, args.channels.split(",")))

    scientisst = ScientISST(address, log=args.log)

    try:
        if args.output:
            firmware_version = scientisst.version_and_adc_chars(print=False)
            file_writer = FileWriter(
                args.output,
                address,
                args.fs,
                args.channels,
                args.convert,
                VERSION,
                firmware_version,
            )
        if args.stream:
            from sense_src.stream_lsl import StreamLSL

            lsl = StreamLSL(
                args.channels,
                args.fs,
                address,
            )
        if args.script:
            script = get_custom_script(args.script)

        stop_event = Event()

        scientisst.start(args.fs, args.channels)
        sys.stdout.write("Start acquisition\n")

        if args.output:
            file_writer.start()
        if args.stream:
            lsl.start()
        if args.script:
            script.start()

        timer = None
        if args.duration > 0:
            timer = run_scheduled_task(args.duration, stop_event)
        try:
            if args.verbose:
                header = "\t".join(get_header(args.channels, args.convert)) + "\n"
                sys.stdout.write(header)
            while not stop_event.is_set():
                frames = scientisst.read(convert=args.convert)
                if args.output:
                    file_writer.put(frames)
                if args.stream:
                    lsl.put(frames)
                if args.script:
                    script.put(frames)
                if args.verbose:
                    sys.stdout.write("{}\n".format(frames[0]))
        except KeyboardInterrupt:
            if args.duration and timer:
                timer.cancel()
            pass

        scientisst.stop()
        # let the acquisition stop before stoping other threads
        time.sleep(0.25)

        sys.stdout.write("Stop acquisition\n")
        if args.output:
            file_writer.stop()
        if args.stream:
            lsl.stop()
        if args.script:
            script.stop()

    finally:
        scientisst.disconnect()

    sys.exit(0)


if __name__ == "__main__":
    main()



# import pylsl
from pylsl import StreamInfo, StreamOutlet, local_clock
import sys

from sense_src.thread_builder import ThreadBuilder


class StreamLSL(ThreadBuilder):
    def __init__(self, channels, fs, address):
        super().__init__()
        self.info = StreamInfo(
            "ScientISST Sense",
            "RAW",
            len(channels),
            fs,
            "int32",
            address,
        )

    def start(self):
        # make outlet
        self.outlet = StreamOutlet(self.info)

        self.timestamp = local_clock()
        self.previous_index = -1
        self.dt = 1 / self.info.nominal_srate()

        sys.stdout.write("Start LSL stream\n")

        super().start()

    def thread_method(self, frames):
        chunk = [frame.a for frame in frames]
        num_frames = len(chunk)

        current_index = frames[-1].seq
        lost_frames = current_index - ((self.previous_index + num_frames) & 15)

        if lost_frames > 0:
            self.timestamp = local_clock()
        else:
            self.timestamp += num_frames * self.dt

        self.previous_index = current_index
        self.outlet.push_chunk(chunk, self.timestamp)
